/***
 * Juliac
 * Copyright (C) 2009-2011 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.opt.ultracomp;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.type.InterfaceType;

import org.objectweb.fractal.julia.ComponentInterface;
import org.objectweb.fractal.julia.Interceptor;

/**
 * Injects bound component references into client interface implementations.
 * 
 * The code of this class originates from {@link
 * org.objectweb.fractal.julia.control.binding.InterceptorBindingMixin}. In the
 * context of the ultra-comp mode, we no longer have any interceptor objects
 * which would be referenced by the client interfaces as this is the case with
 * Julia. Instead, interceptors are components which are always included in the
 * control membrane.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.2.4
 */
public abstract class InterceptorBindingMixin implements BindingController {

  // -------------------------------------------------------------------------
  // Private constructor
  // -------------------------------------------------------------------------

  private InterceptorBindingMixin () {
  }

  // -------------------------------------------------------------------------
  // Fields and methods added and overriden by the mixin class
  // -------------------------------------------------------------------------

	/**
	 * Returns the names of the client interfaces of the component to which this
	 * interface belongs.
	 * 
	 * @return the names of the client interfaces of the component to which this
	 *         interface belongs.
	 */
	public String[] listFc() {
		
		Object[] itfs = _this_weaveableC.getFcInterfaces();
		
		/*
		 * Compute the size of the array which must be returned by determining
		 * the number of client interfaces which are bound (i.e. which reference
		 * a non null value.)
		 */
		int size = 0;
		for (int i = 0; i < itfs.length; i++) {
			Interface itf = (Interface) itfs[i];
			InterfaceType it = (InterfaceType) itf.getFcItfType();
			if (it.isFcClientItf()) {
				if (itf instanceof ComponentInterface) {
					ComponentInterface ci = (ComponentInterface) itf;
					Object impl = ci.getFcItfImpl();
					if (impl != null) {
						size++;
					}
				}
			}
		}
		
		/*
		 * Retrieve the names of bound client interfaces.
		 */
		String[] names = new String[size];
		for (int i=0,j=0; i < itfs.length; i++) {
			Interface itf = (Interface) itfs[i];
			InterfaceType it = (InterfaceType) itf.getFcItfType();
			if (it.isFcClientItf()) {
				if (itf instanceof ComponentInterface) {
					ComponentInterface ci = (ComponentInterface) itf;
					Object impl = ci.getFcItfImpl();
					if (impl != null) {
						names[j] = itf.getFcItfName();
						j++;
					}
				}
			}
		}
		
		return names;
	}
  
	/**
	 * Returns the interface to which the given client interface is bound. More
	 * precisely, returns the server interface to which the client interface
	 * whose name is given is bound. This server interface is necessarily in the
	 * same address space as the client interface (see {@link #bindFc bindFc}).
	 * 
	 * @param clientItfName
	 *            the name of a client interface of the component to which this
	 *            interface belongs.
	 * @return the server interface to which the given interface is bound, or
	 *         <tt>null</tt> if it is not bound.
	 * @throws NoSuchInterfaceException
	 *             if the component to which this interface belongs does not
	 *             have a client interface whose name is equal to the given
	 *             name.
	 */
	public Object lookupFc(String clientItfName)
	throws NoSuchInterfaceException {
		
		Object o = _this_weaveableC.getFcInterface(clientItfName);
		ComponentInterface ci = (ComponentInterface) o;
		Object impl = ci.getFcItfImpl();
		return impl;
	}

	/**
	 * Gets the interceptor of the given client interface, updates its
	 * {@link Interceptor#getFcItfDelegate getFcItfDelegate}, and then calls the
	 * overriden method with the interceptor as server interface (if there is an
	 * interceptor for the client interface).
	 * 
	 * @param clientItfName
	 *            the name of a client interface of the component to which this
	 *            interface belongs.
	 * @param serverItf
	 *            a server interface.
	 * @throws NoSuchInterfaceException
	 *             if there is no such client interface.
	 * @throws IllegalBindingException
	 *             if the binding cannot be created.
	 * @throws IllegalLifeCycleException
	 *             if this component has a
	 *             {@link org.objectweb.fractal.api.control.LifeCycleController}
	 *             interface, but it is not in an appropriate state to perform
	 *             this operation.
	 */
	public void bindFc(final String clientItfName, final Object serverItf)
	throws
		NoSuchInterfaceException, IllegalBindingException,
		IllegalLifeCycleException {
		
		Object o = _this_weaveableC.getFcInterface(clientItfName);
		if (o instanceof ComponentInterface) {
			((ComponentInterface) o).setFcItfImpl(serverItf);
			o = serverItf;
		} else {
			o = serverItf;
		}
		_super_bindFc(clientItfName, o);
	}

	/**
	 * Gets the interceptor of the given client interface, updates its
	 * {@link Interceptor#getFcItfDelegate getFcItfDelegate}, and then calls the
	 * overriden method (if there is an interceptor for the client interface).
	 * 
	 * @param clientItfName
	 *            the name of a client interface of the component to which this
	 *            interface belongs.
	 * @throws NoSuchInterfaceException
	 *             if there is no such client interface.
	 * @throws IllegalBindingException
	 *             if the binding cannot be removed.
	 * @throws IllegalLifeCycleException
	 *             if this component has a
	 *             {@link org.objectweb.fractal.api.control.LifeCycleController}
	 *             interface, but it is not in an appropriate state to perform
	 *             this operation.
	 */
	public void unbindFc(final String clientItfName)
	throws
		NoSuchInterfaceException, IllegalBindingException,
		IllegalLifeCycleException {
		
		Object o = _this_weaveableC.getFcInterface(clientItfName);
		if (o instanceof ComponentInterface) {
			((ComponentInterface) o).setFcItfImpl(null);
		}
		_super_unbindFc(clientItfName);
	}

	// -------------------------------------------------------------------------
	// Fields and methods required by the mixin class in the base class
	// -------------------------------------------------------------------------

	/**
	 * The <tt>weaveableC</tt> field required by this mixin. This field is
	 * supposed to reference the {@link Component} interface of the component to
	 * which this controller object belongs.
	 */
	public Component _this_weaveableC;

	/**
	 * The {@link BindingController#bindFc bindFc} method overriden by this
	 * mixin.
	 * 
	 * @param clientItfName
	 *            the name of a client interface of the component to which this
	 *            interface belongs.
	 * @param serverItf
	 *            a server interface.
	 * @throws NoSuchInterfaceException
	 *             if there is no such client interface.
	 * @throws IllegalBindingException
	 *             if the binding cannot be created.
	 * @throws IllegalLifeCycleException
	 *             if this component has a
	 *             {@link org.objectweb.fractal.api.control.LifeCycleController}
	 *             interface, but it is not in an appropriate state to perform
	 *             this operation.
	 */
	public abstract void _super_bindFc(String clientItfName, Object serverItf)
	throws
		NoSuchInterfaceException, IllegalBindingException,
		IllegalLifeCycleException;

	/**
	 * The {@link BindingController#unbindFc unbindFc} method overriden by this
	 * mixin.
	 * 
	 * @param clientItfName
	 *            the name of a client interface of the component to which this
	 *            interface belongs.
	 * @throws NoSuchInterfaceException
	 *             if there is no such client interface.
	 * @throws IllegalBindingException
	 *             if the binding cannot be removed.
	 * @throws IllegalLifeCycleException
	 *             if this component has a
	 *             {@link org.objectweb.fractal.api.control.LifeCycleController}
	 *             interface, but it is not in an appropriate state to perform
	 *             this operation.
	 */
	public abstract void _super_unbindFc(String clientItfName)
	throws
		NoSuchInterfaceException, IllegalBindingException,
		IllegalLifeCycleException;
}
